import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import javax.swing.JFrame;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;

/**
 * A serial Java program which displays a Mandelbrot image in a window. The
 * portion of the complex plane that is displayed is determined by the points
 * (RL_MIN, IM_MIN) and (RL_MAX, IM_MAX), which represent the lower-left and
 * upper-right points, respectively, of the complex plane.
 * 
 * The width and height of the window are determined by WIDTH and HEIGHT,
 * respectively.
 * 
 * This program borrows from: 1) http://rosettacode.org/wiki/Mandelbrot_set#Java
 * 2) http://phoenix.goucher.edu/~kelliher/s2012/cs325/mandelbrot.c
 * 
 * The coloring rule is the "Close" option described at
 * http://www.math.utah.edu/~pa/math/mandelbrot/mandelbrot.html#color
 * 
 * Although this code is serial, it already contains the necessary "hooks" for
 * parallelizing it.
 * 
 * @author Tom Kelliher
 * 
 */

@SuppressWarnings("serial")
public class Mandelbrot extends JFrame {

	// Number of worker threads.
	private final int threadCount = 1;
	// Array to hold worker thread references.
	private final Thread[] threads = new Thread[threadCount];

	// Maximum number of sequence elements to generate. For a given complex
	// point, if we generate this many sequence elements, we declare the point
	// to be in the Mandelbrot Set.
	private final int MAX_ITER = 100;

	// Width and height of the window.
	private final int WIDTH = 750;
	private final int HEIGHT = 600;

	// Bounds of the complex plane under consideration.
	private final double RL_MIN = -2.2;
	private final double IM_MIN = -1.2;
	private final double RL_MAX = 0.8;
	private final double IM_MAX = 1.2;

	// Diameter squared of the disk within which sequence elements must remain.
	private final double D_SQ = 5.0;

	// Pixels corresponding to complex points in the Mandelbrot set are colored
	// this color.
	private final Color MANDELBROT = new Color(0, 0, 0);
	// Pixels corresponding to complex points that "almost" made it in to the
	// Mandelbrot set are colored this color.
	private final Color CLOSE = new Color(255, 255, 255);
	// Pixels corresponding to complex points whose sequences quickly left the
	// disk are colored this color.
	private final Color FAR = new Color(20, 50, 125);

	// The image into which the Mandelbrot Set image is drawn.
	private BufferedImage I;

	/**
	 * Create the window and construct the Mandelbrot Set image.
	 */
	public Mandelbrot() {
		super("Mandelbrot Set");
		setBounds(0, 0, WIDTH, HEIGHT);
		setResizable(false);
		setDefaultCloseOperation(EXIT_ON_CLOSE);
		I = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_INT_RGB);

		// Create runnable objects for computing the pixel colors for the Mandelbrot
		// Set image, bind the runnables to thread objects, and start the threads.
		for (int i = 0; i < threadCount; i++) {
			Runnable r = worker(getHeight(), getWidth());
			threads[i] = new Thread(r);
			threads[i].start();
		}

		// Make the main thread wait for all of the worker threads to finish before
		// calling for the final image to be redrawn.
		for (int i = 0; i < threadCount; i++) {
			try {
				threads[i].join();
			} catch (InterruptedException e) {
			}
		}

		repaint();
	}

	/*
	 * Some pixel colors take longer to compute than other pixel colors. One way to
	 * parallelize this code would be to split the image into threadCnt horizontal
	 * strips, assigning one worker thread to each strip. However, this could leave
	 * one worker thread with all the computationally intensive pixel colors to
	 * compute.
	 * 
	 * We seek an approach that balances the computational work among the threads.
	 * 
	 * The approach to use is to have each worker thread work on one row at a time.
	 * When a worker finishes one row, it gets another row to work on. This starts
	 * with row 0 and continues to row (getHeight() - 1).
	 * 
	 * To implement this approach, you will need to modify worker(). The outer for
	 * loop should be transformed into a while loop. (I.e., while there are more
	 * rows to color...) Add a new private int field, nextRow, to keep track of the
	 * next row to be colored. To avoid synchronization problems, access to the
	 * shared variables (nextRow and I), must be guarded by using a ReentrantLock.
	 * To maximize concurrency, be careful to minimize the amount of code guarded by
	 * the ReentrantLock. To further increase concurrency, you could use two
	 * ReentrantLocks.
	 * 
	 * To test your parallel code, set threadCount to 4. Add a few println() calls
	 * to demonstrate that your main thread and all of the worker threads are
	 * terminating.
	 */

	private Runnable worker(int rows, int width) {
		return () -> {
			// Assign colors to the pixels of I. The outer loop iterates
			// over the image's rows. The inner loop iterates over a row's
			// pixels. Rows are numbered from top to bottom. Within a row,
			// pixels are numbered from left to right. I.E., the upper left
			// corner of the image has coordinates (0, 0).
			for (int y = 0; y < rows; y++) {
				for (int x = 0; x < width; x++) {
					Point2D.Double complex = windowToComplex(new Point(x, y));
					Color c = colorPixel(complex);
					I.setRGB(x, y, c.getRGB());
				}
			}
		};
	}

	/**
	 * @param c A point in the complex plane.
	 * @return The color with which to color the pixel corresponding to c.
	 */
	private Color colorPixel(Point2D.Double c) {
		double cX = c.x;
		double cY = c.y;
		double zX = 0.0;
		double zY = 0.0;
		double tmp;
		int iter = 0;
		// Generate sequence elements and check that they remain within the disk.
		while (zX * zX + zY * zY < D_SQ && iter < MAX_ITER) {
			tmp = zX * zX - zY * zY + cX;
			zY = 2.0 * zX * zY + cY;
			zX = tmp;
			iter++;
		}
		if (iter == MAX_ITER)
			// This point is in the Mandelbrot set.
			return MANDELBROT;
		else {
			// This point isn't in the Mandelbrot set. Interpolate a color.
			double a = (double) iter / MAX_ITER;
			a = Math.sqrt(a);
			return lerp(a, CLOSE, FAR);
		}
	}

	/**
	 * Map from a pixel in the window to a point in the complex plane. This method
	 * recognizes that the window uses a left-handed coordinate system and that the
	 * complex plane is a right-handed coordinate system.
	 * 
	 * @param pixel A particular pixel in the window.
	 * @return The point in the complex plane corresponding to the pixel.
	 */
	private Point2D.Double windowToComplex(Point pixel) {
		double rlDelta = (RL_MAX - RL_MIN) / getWidth();
		double imDelta = (IM_MAX - IM_MIN) / getHeight();
		double rl = RL_MIN + pixel.x * rlDelta;
		double im = IM_MAX - pixel.y * imDelta;
		return new Point2D.Double(rl, im);
	}

	/**
	 * Linearly interpolate between two colors.
	 * 
	 * @param alpha The linear interpolation coefficient.
	 * @param a     The first color.
	 * @param b     The second color.
	 * @return The interpolated color.
	 */
	private Color lerp(double alpha, Color a, Color b) {
		assert (0.0 <= alpha && alpha <= 1.0);
		int red = lerp(alpha, a.getRed(), b.getRed());
		int green = lerp(alpha, a.getGreen(), b.getGreen());
		int blue = lerp(alpha, a.getBlue(), b.getBlue());
		return new Color(red, green, blue);
	}

	/**
	 * Linearly interpolate between two integers.
	 * 
	 * @param alpha The linear interpolation coefficient.
	 * @param a     The first integer.
	 * @param b     The second integer.
	 * @return The interpolated integer.
	 */
	private int lerp(double alpha, int a, int b) {
		assert (0.0 <= alpha && alpha <= 1.0);
		return (int) (alpha * a + (1.0 - alpha) * b);
	}

	@Override
	public void paint(Graphics g) {
		g.drawImage(I, 0, 0, this);
	}

	public static void main(String[] args) {
		new Mandelbrot().setVisible(true);
	}
}